import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import optimizers, regularizers
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau
from tensorflow.keras.layers import Input, Dense, Activation, BatchNormalization, Dropout, Lambda, Concatenate, concatenate
from tensorflow.keras.models import Model
from tensorflow.keras.losses import mse
from tensorflow.keras.utils import to_categorical
from sklearn.utils import shuffle
import numpy as np
from tensorflow.python.keras.layers import Layer
import anndata
import scbean.tools.utils as tl


def sampling(args):
    z_mean, z_log_var = args
    batch = K.shape(z_mean)[0]
    dim = K.int_shape(z_mean)[1]
    epsilon = K.random_normal(shape=(batch, dim))
    return z_mean + K.exp(0.5 * z_log_var) * epsilon

#
# def reverse_gradient(X, hp_lambda):
#     '''Flips the sign of the incoming gradient during training.'''
#     try:
#         reverse_gradient.num_calls += 1
#     except AttributeError:
#         reverse_gradient.num_calls = 1
#
#     grad_name = "GradientReversal%d" % reverse_gradient.num_calls
#
#     @tf.RegisterGradient(grad_name)
#     def _flip_gradients(op, grad):
#         return [tf.negative(grad) * hp_lambda]
#
#     g = tf.compat.v1.get_default_graph()
#     with g.gradient_override_map({'Identity': grad_name}):
#         y = tf.identity(X)
#
#     return y
#
#
# class GradientReversal(Layer):
#     '''Flip the sign of gradient during training.'''
#     def __init__(self, hp_lambda, **kwargs):
#         super(GradientReversal, self).__init__(**kwargs)
#         self.supports_masking = False
#         self.hp_lambda = hp_lambda
#
#     def build(self, input_shape):
#         self._trainable_weights = []
#
#     def call(self, x, mask=None):
#         return reverse_gradient(x, self.hp_lambda)
#
#     def get_output_shape_for(self, input_shape):
#         return input_shape
#
#     def get_config(self):
#         config = {'hp_lambda': self.hp_lambda}
#         base_config = super(GradientReversal, self).get_config()
#         return dict(list(base_config.items()) + list(config.items()))

@tf.custom_gradient
def grad_reverse(x):
    y = tf.identity(x)
    def custom_grad(dy):
        return -dy
    return y, custom_grad


class GradReverse(tf.keras.layers.Layer):
    def __init__(self):
        super().__init__()

    def call(self, x):
        return grad_reverse(x)


class DAVAE:
    def __init__(self, input_size, batches=2, domain_scale_factor=1.0, 
                 hidden_layers=[128, 64, 32, 5], path=''):
        self.input_size = input_size
        self.path = path
        self.dann_vae = None
        self.inputs = None
        self.outputs_x = None
        self.initializers = "glorot_uniform"
        self.optimizer = optimizers.Adam(lr=0.01)
        self.hidden_layers = hidden_layers
        self.domain_scale_factor = domain_scale_factor
        self.dropout_rate_small = 0.01
        self.dropout_rate_big = 0.05
        self.kernel_regularizer = regularizers.l1_l2(l1=0.00, l2=0.00)
        self.validation_split = 0.0
        self.batches = batches
        self.dropout_rate = 0.01
        callbacks = []
        checkpointer = ModelCheckpoint(filepath=path + "vae_weights.h5", verbose=1, save_best_only=False,
                                       save_weights_only=True)
        reduce_lr = ReduceLROnPlateau(monitor='loss', factor=0.8, patience=100, min_lr=0.0001)
        early_stop = EarlyStopping(monitor='loss', patience=200)
        tensor_board = TensorBoard(log_dir=path + 'logs/')
        callbacks.append(checkpointer)
        callbacks.append(reduce_lr)
        callbacks.append(early_stop)
        callbacks.append(tensor_board)
        self.callbacks = callbacks

    def build(self):
        Relu = "relu"
        en_ly_size = len(self.hidden_layers)
        z_size = self.hidden_layers[en_ly_size - 1]

        inputs_x = Input(shape=(self.input_size,), name='inputs')
        inputs_batch = Input(shape=(self.batches,), name='inputs_batch')
        inputs_loss_weights = Input(shape=(1,), name='inputs_weights')
        x = inputs_x
        for i in range(en_ly_size):
            if i == en_ly_size - 1:
                break
            ns = self.hidden_layers[i]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate)(x)

        hx_mean = Dense(z_size, kernel_regularizer=self.kernel_regularizer,
                        kernel_initializer=self.initializers,
                        name="hx_mean")(x)
        hx_log_var = Dense(z_size, kernel_regularizer=self.kernel_regularizer,
                           kernel_initializer=self.initializers,
                           name="hx_log_var")(x)
        hx_z = Lambda(sampling, output_shape=(z_size,), name='hx_z')([hx_mean, hx_log_var])
        encoder_hx = Model(inputs_x, [hx_mean, hx_log_var, hx_z], name='encoder_hx')

        latent_inputs_x = Input(shape=(z_size,), name='latent')
        x = latent_inputs_x
        for i in range(en_ly_size - 1, 0, -1):
            ns = self.hidden_layers[i - 1]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate_big)(x)

        outputs_x = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                          kernel_initializer=self.initializers, activation="softplus")(x)
        decoder_x = Model(latent_inputs_x, outputs_x, name='decoder_x')

        latent_inputs_batch = Input(shape=(z_size,), name='latent_domain')
        Flip = GradReverse()
        d = Flip(latent_inputs_batch)
        d = Dense(16, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(d)
        d = BatchNormalization(center=True, scale=False)(d)
        d = Activation(Relu)(d)
        d = Dropout(self.dropout_rate_big)(d)

        d = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation="softmax")(d)
        domian_classifier = Model(latent_inputs_batch, d, name='domain_classifier')

        outputs_x = decoder_x(encoder_hx(inputs_x)[2])
        domain_pred = domian_classifier(encoder_hx(inputs_x)[2])

        dann_vae = Model([inputs_x, inputs_batch, inputs_loss_weights], [outputs_x, domain_pred], name='vae_mlp')

        inputs_x = tf.multiply(inputs_x, inputs_loss_weights)
        outputs_x = tf.multiply(outputs_x, inputs_loss_weights)

        reconstruction_loss = mse(inputs_x, outputs_x)
        # reconstruction_loss = mse(inputs_x, outputs_x)

        noise = tf.math.subtract(inputs_x, outputs_x)
        var = tf.math.reduce_variance(noise)
        reconstruction_loss *= (0.5*self.input_size)/var
        reconstruction_loss += (0.5*self.input_size)/var*tf.math.log(var)

        kl_loss_z = -0.5 * K.sum(1 + hx_log_var - K.square(hx_mean) - K.exp(hx_log_var), axis=-1)

        pred_loss = K.categorical_crossentropy(inputs_batch, domain_pred)*self.input_size*self.domain_scale_factor
        vae_loss = K.mean(reconstruction_loss + kl_loss_z + pred_loss)

        dann_vae.add_loss(vae_loss)
        self.dann_vae = dann_vae
        self.encoder = encoder_hx
        self.decoder = decoder_x

    def compile(self):
        self.dann_vae.compile(optimizer=self.optimizer)
        self.dann_vae.summary()

    def train(self, x, batch, loss_weights, batch_size=100, epochs=300):
        history = self.dann_vae.fit({'inputs': x, 'inputs_batch': batch, 'inputs_weights': loss_weights},
                                    epochs=epochs, batch_size=batch_size,
                                    validation_split=self.validation_split, shuffle=True)
        return history

    def get_output(self, x, batches,):
        [z_mean, z_log_var, z] = self.encoder.predict(x)
        output_x = self.decoder.predict(z_mean)
        return z_mean, output_x


class DACVAE:
    def __init__(self, input_size, batches=2, domain_scale_factor=1.0, hidden_layers=[128, 64, 32, 5], path=''):
        self.input_size = input_size
        self.path = path
        self.dann_vae = None
        self.inputs = None
        self.outputs_x = None
        self.initializers = "glorot_uniform"
        self.optimizer = optimizers.Adam(lr=0.01)
        self.hidden_layers = hidden_layers
        self.dropout_rate_small = 0.01
        self.dropout_rate_big = 0.1
        self.kernel_regularizer = regularizers.l1_l2(l1=0.00, l2=0.00)
        self.domain_scale_factor = domain_scale_factor
        self.validation_split = 0.0
        self.batches = batches
        self.dropout_rate = 0.01
        callbacks = []
        checkpointer = ModelCheckpoint(filepath=path + "vae_weights.h5", verbose=1, save_best_only=False,
                                       save_weights_only=True)
        reduce_lr = ReduceLROnPlateau(monitor='loss', factor=0.8, patience=100, min_lr=0.0001)
        early_stop = EarlyStopping(monitor='loss', patience=200)
        tensor_board = TensorBoard(log_dir=path + 'logs/')
        callbacks.append(checkpointer)
        callbacks.append(reduce_lr)
        callbacks.append(early_stop)
        callbacks.append(tensor_board)
        self.callbacks = callbacks

    def build(self):
        Relu = "relu"
        en_ly_size = len(self.hidden_layers)
        z_size = self.hidden_layers[en_ly_size-1]

        inputs_x = Input(shape=(self.input_size,), name='inputs')
        inputs_batch = Input(shape=(self.batches,), name='inputs_batch')
        inputs_loss_weights = Input(shape=(1,), name='inputs_weights')
        inputs = concatenate([inputs_x, inputs_batch])
        x = inputs
        for i in range(en_ly_size):
            if i == en_ly_size - 1:
                break
            ns = self.hidden_layers[i]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate_big)(x)

        hx_mean = Dense(z_size, kernel_regularizer=self.kernel_regularizer,
                        kernel_initializer=self.initializers,
                        name="hx_mean")(x)
        hx_log_var = Dense(z_size, kernel_regularizer=self.kernel_regularizer,
                           kernel_initializer=self.initializers,
                           name="hx_log_var")(x)
        hx_z = Lambda(sampling, output_shape=(z_size,), name='hx_z')([hx_mean, hx_log_var])
        encoder_hx = Model([inputs_x, inputs_batch], [hx_mean, hx_log_var, hx_z], name='encoder_hx')

        latent_inputs_x = Input(shape=(z_size,), name='latent')
        latent_inputs_batch = Input(shape=(self.batches,), name='latent_batch')
        latent_inputs = concatenate([latent_inputs_x, latent_inputs_batch])
        x=latent_inputs
        for i in range(en_ly_size - 1, 0, -1):
            ns = self.hidden_layers[i - 1]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate_big)(x)

        outputs_x = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                          kernel_initializer=self.initializers, activation="softplus")(x)
        decoder_x = Model([latent_inputs_x, latent_inputs_batch], outputs_x, name='decoder_x')

        latent_inputs_domain = Input(shape=(z_size,), name='latent_domain')
        Flip = GradReverse()
        # d= grad_reverse(latent_inputs_batch)
        d = Flip(latent_inputs_domain)
        d = Dense(16, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(d)
        d = BatchNormalization(center=True, scale=False)(d)
        d = Activation(Relu)(d)
        d = Dropout(self.dropout_rate_small)(d)

        d = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation="softmax")(d)
        domian_classifier = Model(latent_inputs_domain, d, name='domain_classifier')

        outputs_x = decoder_x([encoder_hx([inputs_x, inputs_batch])[2], inputs_batch])
        domain_pred = domian_classifier(encoder_hx([inputs_x, inputs_batch])[2])

        dann_vae = Model([inputs_x, inputs_batch, inputs_loss_weights], [outputs_x, domain_pred], name='vae_mlp')

        inputs_x = tf.multiply(inputs_x, inputs_loss_weights)
        outputs_x = tf.multiply(outputs_x, inputs_loss_weights)
        reconstruction_loss = mse(inputs_x, outputs_x)
        # hx_log_var = tf.multiply(hx_log_var, inputs_loss_weights)
        # hx_mean = tf.multiply(hx_mean, inputs_loss_weights)
        # reconstruction_loss = mse(inputs_x, outputs_x)

        noise = tf.math.subtract(inputs_x, outputs_x)
        var = tf.math.reduce_variance(noise)
        reconstruction_loss *= (0.5*self.input_size)/var
        reconstruction_loss += (0.5*self.input_size)/var*tf.math.log(var)

        kl_loss_z = -0.5 * K.sum(1 + hx_log_var - K.square(hx_mean) - K.exp(hx_log_var), axis=-1)

        pred_loss = K.categorical_crossentropy(inputs_batch, domain_pred)*self.input_size*self.domain_scale_factor
        vae_loss = K.mean(reconstruction_loss + kl_loss_z + pred_loss)

        dann_vae.add_loss(vae_loss)
        self.dann_vae = dann_vae
        self.encoder = encoder_hx
        self.decoder = decoder_x

    def compile(self):
        self.dann_vae.compile(optimizer=self.optimizer)
        self.dann_vae.summary()

    def train(self, x, batch, loss_weights, batch_size=100, epochs=300):

        history = self.dann_vae.fit({'inputs': x, 'inputs_batch': batch, 'inputs_weights': loss_weights},
                                        epochs=epochs, batch_size=batch_size,
                                        validation_split=self.validation_split, shuffle=True)
        return history

    def get_output(self, x, batches,):
        [z_mean, z_log_var, z] = self.encoder.predict([x, batches])
        output_x = self.decoder.predict([z_mean, batches])
        return z_mean, output_x




def fit_integration(adata, batch_num=2, mode='DACVAE', split_by='batch_label', epochs=20, batch_size=128,
                    domain_lambda=1.0, sparse=True, hidden_layers=[128,64,32,5]):
    """/
    Build DAVAE model and fit the data to the model for training.

    Parameters
    ----------
    adata: AnnData
        AnnData object need to be integrated.

    batch_num: int, optional (default: 2)
        Number of batches of datasets to be integrated.

    mode: string, optional (default: 'DACVAE')
        if 'DACVAE', construct a DACVAE model
        if 'DAVAE', construct a DAVAE model

    split_by: string, optional (default: '_batch')
        the obsm_name of obsm used to distinguish different batches.

    epochs: int, optional (default: 200)
        Number of epochs to train the model. An epoch is an iteration over the entire x and y data provided.

    batch_size: int or None, optional (default: 256)
        Number of samples per gradient update. If unspecified, batch_size will default to 32.

    domain_lambda: double, optional (default: 1.0)
        The coefficient multiplied by the loss value of the domian classifier of DAVAE model.

    sparse: bool, optional (default: True)
        If True, Matrix X in the AnnData object is stored as a sparse matrix.

    hidden_layers: list of integers, (default: [128,64,32,5])
        Number of hidden layer neurons in the model.

    Returns
    -------
    :class:`~anndata.AnnData`
        out_adata
    """

    batch = adata.obs[split_by]
    batch = np.array(batch.values, dtype=int)
    loss_weight = adata.obs['loss_weight']
    # orig_batch, batch_num = tl.generate_batch_code(batch, batch_num)
    orig_batch  = to_categorical(batch)
    if sparse:
        orig_data = adata.X.A
    else:
        orig_data = adata.X
    data, batch, loss_weight = shuffle(orig_data, orig_batch, loss_weight, random_state=0)
    if mode=='DAVAE':
        net = DAVAE(input_size=data.shape[1], batches=batch_num, domain_scale_factor=domain_lambda,
                    hidden_layers=hidden_layers)
    else:
        net = DACVAE(input_size=data.shape[1],  batches=batch_num, domain_scale_factor=domain_lambda,
                     hidden_layers=hidden_layers)
    net.build()
    net.compile()
    net.train(data, batch, loss_weight, batch_size=batch_size, epochs=epochs)
    latent_z, output_x = net.get_output(orig_data, orig_batch)
    out_adata = anndata.AnnData(X=output_x, obs=adata.obs, var=adata.var)
    out_adata.obsm['X_davae'] = latent_z
    out_adata.raw = adata.copy()
    return out_adata




import numpy as np
import math
import scanpy as sc
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
matplotlib.use('Agg')
sns.set(style='white', rc={'figure.figsize':(5,5), 'figure.dpi':150})


def findNeighbors(adata):
	sc.pp.neighbors(adata, n_neighbors=50,use_rep='X_vipcca',n_pcs=16)
	ann=adata.obs
	ann['nid']=range(ann.shape[0])
	ct=ann[ann['tech']=="rna"]['celltype'].unique()
	ann_atac=ann[ann['tech']=="atac"]
	ann_rna=ann[ann['tech']=="rna"]
	weights=ann_rna['celltype'].value_counts()/ann_rna.shape[0]
	for item in ct:
		ann_atac[item]=0.0
	ann_atac["pred_max_score"]=0.0
	querycells = ann[ann['tech']=="atac"]['nid']
	dist=adata.uns['neighbors']['distances']
	weights[weights.index]=np.ones(len(weights))
	weights["Platelets"]=0.001
	
	for cell in querycells:
		cell_key=ann.index[cell]
		nbs=dist.getrow(cell).nonzero()[1]
		ann_nbs=ann.iloc[nbs,:]
		ann_nbs=ann_nbs[ann_nbs['tech']=='rna']
		if ann_nbs.empty is not True:
			for key in ann_nbs.index:
				nbid=ann_nbs.loc[key,'nid']
				ct=ann_nbs.at[key,'celltype']
				var=weights[ann_nbs.at[key,'celltype']]**2
				ann_atac.at[cell_key,ct]+=math.exp(-(dist[cell,nbid].tolist()**2)/(2*var))
			ann_atac.at[cell_key,'pred_max_score'] = max(ann_atac.loc[cell_key,][9:22])
	ann_atac.loc[:,'celltype']=ann_atac.iloc[:,9:22].idxmax(axis=1)
	return ann_atac


def findNeighbors2(adata):
	sc.pp.neighbors(adata, n_neighbors=50,use_rep='X_vipcca',n_pcs=16)
	ann=adata.obs
	ann['nid']=range(ann.shape[0])
	ct=ann[ann['tech']=="rna"]['celltype'].unique()
	ann_atac=ann[ann['tech']=="atac"]
	ann_rna=ann[ann['tech']=="rna"]
	querycells = ann[ann['tech']=="atac"]['nid']
	dist=adata.uns['neighbors']['distances']
	for cell in querycells:
		cell_key=ann.index[cell]
		if cell_key == "TCACAGAGTAACGGAC-1":
			print(cell_key)
		nbs=dist.getrow(cell).nonzero()[1]
		ann_nbs=ann.iloc[nbs,:]
		ann_nbs=ann_nbs[ann_nbs['tech']=='rna']
		if ann_nbs.empty is not True:
				ann_atac.at[cell_key,'celltype']=ann_nbs.celltype.value_counts().idxmax()
	return ann_atac



import argparse

def parse():
	parser = argparse.ArgumentParser(description = "This tool is used to analyse single cell data.")
	parser.add_argument('--version', action='version', version='%(prog)s 0.1')
	
	subparsers = parser.add_subparsers(dest='command',help='commands')
	train_parser = subparsers.add_parser('train', help='train contents')
	train_parser.add_argument('--ref_file','-r', required = True, 
					help='The path to the reference file.')
	train_parser.add_argument('--epochs','-e', type=int, default=3000,
					help='The size of each batch.')
	train_parser.add_argument('--patience','-pt', type=int, default=50,
					help='The size of each batch.')
	train_parser.add_argument('--name_ref','-n', default = "", 
					help='The name of the reference.')
	train_parser.add_argument('--batch_size','-b', type=int, default=128,
					help='The size of each batch.')
					
	predict_parser = subparsers.add_parser('predict', help='predict contents')
	predict_parser.add_argument('--query_file','-q', required = True, 
					help='The path to one or multiple query file.')
	predict_parser.add_argument('--groups','-g', default = "", 
					help='the columns containing group information, e.g. cell types and batches.',type=str)
	
	parser.add_argument('--path_to_result','-p', default = "./test_results/", 
					help='The path where the result files should be written.')
	parser.add_argument('--format','-f', default= "h5ad", 
					choices=['h5ad','10x_mtx','10x_h5'],
					help='The format of input files. It has two options, \'h5ad\' or \'10x_mtx\', or \'10x_h5\'.')
	parser.add_argument('--y_size','-y', type=int, default=10,
					help='The size of vector y.')
	parser.add_argument('--scale','-s', type=bool, default=True,
					help='The size of each batch.')
	parser.add_argument('--min_counts','-mc', type=int, default=50,
					help='The size of each batch.')
	parser.add_argument('--min_genes','-mg', type=int, default=200,
					help='The size of each batch.')
	args = parser.parse_args()
	return args





from scipy.stats import norm
from scipy.stats import rankdata
import numpy as np

def cacu_theory_quantiles(data, vx, scale=1):
    if vx > 10:
        sx = (data-0.5)/vx
    else:
        sx = (data-0.375)/(vx+0.25)
    sy = norm.ppf(sx, scale=scale)
    return sy

def rank(X, method="average", axis=0):
    rm=np.ndarray(shape=X.shape)
    for i in range(X.shape[1-axis]):
        if axis==0:
            data = X[:,i]
            rm[:,i]=rankdata(data, method=method)
        else:
            data = X[i,:]
            rm[i,:]=rankdata(data, method=method)
    return rm

def qqnorm(X, method='average', axis=0, scale=1):
    rm=rank(X, method=method, axis=axis)
    n_quantiles=rm.max(axis=axis)
    X_new = np.ndarray(shape=X.shape)
    for i in range(len(n_quantiles)):
        n_quantile=n_quantiles[i]
        if axis==0:
            X_new[:,i]=cacu_theory_quantiles(rm[:,i],n_quantile, scale=scale)
        else:
            X_new[i,:]=cacu_theory_quantiles(rm[i,:],n_quantile, scale=scale)
    return X_new



from .utils import generate_datasets,generate_img_from_genes, read_sc_data, write2mtx, createCoordination, add_location, logNormalization, recipe_vipcca, preprocessing, split_object, scale, rank_normalization, concatenate_datasets, davae_preprocessing, generate_batch_code, spatial_preprocessing, spatial_rna_preprocessing
from .qqnorm import qqnorm
from .parser import parse
from .plotting import run_embedding, plotEmbedding, plotDEG, plotDEG2, runGeoSketch, plotPrediction, plotCorrelation
from .transferLabel import findNeighbors, findNeighbors2


import tensorflow as tf
from tensorflow.keras.models import Model, load_model
from tensorflow.keras import backend as K
from tensorflow.keras.losses import mse, binary_crossentropy
from tensorflow.keras import optimizers, regularizers, initializers
from tensorflow.keras.layers import Input, Dense, Activation, BatchNormalization, Dropout, Lambda, Concatenate, concatenate
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau
import os
import os.path

import matplotlib.pyplot as plt
import numpy as np


def sampling(args):
    z_mean, z_log_var = args
    epsilon = K.random_normal(shape=K.shape(z_mean))
    return z_mean + K.exp(0.5 * z_log_var) * epsilon


# MeanAct = lambda x: tf.clip_by_value(K.exp(x), 1e-5, 1e5)


class VIMCCA:
    def __init__(self, input_size_x, inputs_size_y,
                 hidden_layers,
                 path="./",
                 dropout_rate_big=0.05,
                 dropout_rate_small=0.0,
                 validation_split=0.0,
                 patience=60,
                 weight = 5,
                 deterministic=False):
        self.input_size_x = input_size_x
        self.input_size_y = inputs_size_y
        self.hidden_layers = hidden_layers
        self.dropout_rate_big = dropout_rate_big
        self.dropout_rate_small = dropout_rate_small
        self.weight = weight
        self.VIMCCA = None
        self.inputs = None
        self.outputs = None
        self.path = path
        self.initializers = "glorot_uniform"
        self.optimizer = optimizers.Adam(beta_1=0.9, beta_2=0.999, epsilon=10e-8)
        self.kernel_regularizer = regularizers.l1_l2(l1=0.01, l2=0.01)
        self.validation_split = validation_split
        self.deterministic = deterministic
        callbacks = []
        checkpointer = ModelCheckpoint(filepath=path + "vae_weights.h5", verbose=1, save_best_only=False,
                                       save_weights_only=True)
        reduce_lr = ReduceLROnPlateau(monitor='loss', factor=0.8, patience=10, min_lr=0.0001)
        early_stop = EarlyStopping(monitor='loss', patience=patience)
        tensor_board = TensorBoard(log_dir=path + 'logs/')
        callbacks.append(checkpointer)
        callbacks.append(reduce_lr)
        callbacks.append(early_stop)
        callbacks.append(tensor_board)
        self.callbacks = callbacks

    def build(self):
        en_ly_size = len(self.hidden_layers)
        Relu = "relu"
        inputs_x = Input(shape=(self.input_size_x,), name='inputs_x')
        inputs_y = Input(shape=(self.input_size_y,), name='inputs_y')
        z = inputs_x
        for i in range(en_ly_size):
            if i == en_ly_size - 1:
                break
            ns = self.hidden_layers[i]
            z = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(z)
            z = BatchNormalization(center=True, scale=False)(z)
            z = Activation(Relu)(z)
            z = Dropout(self.dropout_rate_small)(z)
        latent_z_size = self.hidden_layers[i]

        z_mean = Dense(latent_z_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                        bias_initializer='zeros', name="z_mean")(z)
        z_log_var = Dense(latent_z_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                           bias_initializer='zeros', name="z_log_var")(z)
        z = Lambda(sampling, output_shape=(latent_z_size,), name='z')([z_mean, z_log_var])
        encoder_z = Model(inputs_x, [z_mean, z_log_var, z], name='encoder_z')

        latent_inputs_x = Input(shape=(latent_z_size,), name='x')
        latent_inputs_y = Input(shape=(latent_z_size,), name='y')
        x = latent_inputs_x
        y = latent_inputs_y
        for i in range(en_ly_size - 1, 0, -1):
            ns = self.hidden_layers[i - 1]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate_big)(x)

            y = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(y)
            y = BatchNormalization(center=True, scale=False)(y)
            y = Activation(Relu)(y)
            y = Dropout(self.dropout_rate_big)(y)

        outputs_x = Dense(self.input_size_x, kernel_regularizer=self.kernel_regularizer,
                        kernel_initializer=self.initializers, activation="linear")(x)
        decoder_x = Model(latent_inputs_x, outputs_x, name='decoder_x')

        outputs_y = Dense(self.input_size_y, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="linear", name='outputs_y')(y)
        decoder_y = Model(latent_inputs_y, outputs_y, name='decoder_y')

        outputs_x = decoder_x(encoder_z(inputs_x)[2])
        outputs_y = decoder_y(encoder_z(inputs_x)[2])
        VIMCCA = Model([inputs_x, inputs_y], [outputs_x, outputs_y], name='VIMCCA_mlp')

        reconstruction_loss_x = mse(inputs_x, outputs_x)
        reconstruction_loss_y = mse(inputs_y, outputs_y)
        # reconstruction_loss_x = K.sum(binary_crossentropy(inputs_x, outputs_x))
        # reconstruction_loss_y = K.sum(binary_crossentropy(inputs_y, outputs_y))
        # reconstruction_loss_x*=self.input_size_x
        # reconstruction_loss_y*=self.input_size_y
        noise_x = tf.math.subtract(inputs_x, outputs_x)
        var_x = tf.math.reduce_variance(noise_x)
        reconstruction_loss_x *= (0.5 * self.input_size_x) / var_x
        reconstruction_loss_x += (0.5 * self.input_size_x) / var_x * tf.math.log(var_x)
        #
        noise_y = tf.math.subtract(inputs_y, outputs_y)
        var_y = tf.math.reduce_variance(noise_y)
        reconstruction_loss_y *= (0.5 * self.input_size_y) / var_y
        reconstruction_loss_y += (0.5 * self.input_size_y) / var_y * tf.math.log(var_y)
        # reconstruction_loss_y*=5
        # reconstruction_loss_x /= 2000
        # reconstruction_loss_y
        # reconstruction_loss_x = mse(inputs_x, outputs_x)*2000
        # reconstruction_loss_y = mse(inputs_y, outputs_y)*2000
        kl_loss_z = -0.5 * K.sum(1 + z_log_var - K.square(z_mean) - K.exp(z_log_var), axis=-1)
        # reconstruction_loss_x = tf.reduce_mean(reconstruction_loss_x)
        # reconstruction_loss_y = tf.reduce_mean(reconstruction_loss_y)
        # kl_loss_z/=self.hidden_layers[-1]

        vae_loss = K.mean(reconstruction_loss_x + self.weight*reconstruction_loss_y + kl_loss_z)

        VIMCCA.add_loss(vae_loss)
        self.VIMCCA = VIMCCA
        self.encoder_z = encoder_z
        self.decoder_x = decoder_x
        self.decoder_y = decoder_y

    def compile(self):
        self.VIMCCA.compile(optimizer=self.optimizer)
        self.VIMCCA.summary()

    def train(self, x, y, batch_size=1, epochs=300):
        # if os.path.isfile(self.path + "vae_weights.h5"):
        #     self.vae.load_weights(self.path + "vae_weights.h5")
        # else:
        #
        history = self.VIMCCA.fit({'inputs_x': x, 'inputs_y': y}, epochs=epochs, batch_size=batch_size,
                         validation_split=self.validation_split, shuffle=True)

        return history

    def integrate(self, x):
        [z_mean, z_log_var, z] = self.encoder_z.predict(x)
        return z_mean

    def integrate_compose(self, x):
        [z_mean, z_log_var, z] = self.encoder_z.predict(x)
        return z_mean

    def get_output(self, x):
        [z_mean, z_log_var, z] = self.encoder_z.predict(x)
        output_x = self.decoder_x.predict(z_mean)
        output_y = self.decoder_y.predict(z_mean)
        return output_x, output_y

    def plot_loss(self, history):
        print(len(history.history['loss']))
        plt.plot(history.history['loss'])
        plt.show()


# class PVIMCCA:
#     def __init__(self, input_size_x, inputs_size_y,
#                  hidden_layers=[128,64,32,5],
#                  dropout_rate_small=0.01,
#                  dropout_rate_big=0.05,
#                  path="./",
#                  latent_xy_size=2,
#                  batch_num=2,
#                  validation_split=0.0,
#                  patience=60,
#                  deterministic=False):
#         self.input_size_x = input_size_x
#         self.input_size_y = inputs_size_y
#         self.VIMCCA = None
#         self.inputs = None
#         self.outputs = None
#         self.path = path
#         self.initializers = "glorot_normal"
#         self.optimizer = optimizers.Adam(beta_1=0.9, beta_2=0.999, epsilon=10e-8)
#         self.hidden_layers = hidden_layers
#         self.latent_xy_size = latent_xy_size
#         self.dropout_rate_small = dropout_rate_small
#         self.dropout_rate_big = dropout_rate_big
#         self.kernel_regularizer = regularizers.l1_l2(l1=0.01, l2=0.01)
#         self.validation_split = validation_split
#         self.deterministic = deterministic
#
#         self.batch_num = batch_num
#         callbacks = []
#         checkpointer = ModelCheckpoint(filepath=path + "vae_weights.h5", verbose=1, save_best_only=False,
#                                        save_weights_only=True)
#         reduce_lr = ReduceLROnPlateau(monitor='loss', factor=0.8, patience=10, min_lr=0.0001)
#         early_stop = EarlyStopping(monitor='loss', patience=patience)
#         tensor_board = TensorBoard(log_dir=path + 'logs/')
#         callbacks.append(checkpointer)
#         callbacks.append(reduce_lr)
#         callbacks.append(early_stop)
#         callbacks.append(tensor_board)
#         self.callbacks = callbacks
#
#     def build(self):
#         en_ly_size = len(self.hidden_layers)
#         Relu = "relu"
#         inputs_x = Input(shape=(self.input_size_x,), name='inputs_x')
#         inputs_y = Input(shape=(self.input_size_y,), name='inputs_y')
#
#         hx = inputs_x
#         hy = inputs_y
#         for i in range(en_ly_size):
#             if i == en_ly_size - 1:
#                 break
#             ns = self.hidden_layers[i]
#             hx = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(hx)
#             hx = BatchNormalization(center=True, scale=False)(hx)
#             hx = Activation(Relu)(hx)
#             hx = Dropout(self.dropout_rate_small)(hx)
#
#             hy = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(hy)
#             hy = BatchNormalization(center=True, scale=False)(hy)
#             hy = Activation(Relu)(hy)
#             hy = Dropout(self.dropout_rate_small)(hy)
#
#         hx_mean = Dense(self.latent_xy_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                        name="hx_mean")(hx)
#         hx_log_var = Dense(self.latent_xy_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                           name="hx_log_var")(hx)
#         hx_z = Lambda(sampling, output_shape=(self.latent_xy_size,), name='hx_z')([hx_mean, hx_log_var])
#         encoder_hx = Model(inputs_x, [hx_mean, hx_log_var, hx_z], name='encoder_hx')
#
#         hy_mean = Dense(self.latent_xy_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                         name="hy_mean")(hy)
#         hy_log_var = Dense(self.latent_xy_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                            name="hy_log_var")(hy)
#         hy_z = Lambda(sampling, output_shape=(self.latent_xy_size,), name='hy_z')([hy_mean, hy_log_var])
#         encoder_hy = Model(inputs_y, [hy_mean, hy_log_var, hy_z], name='encoder_hy')
#
#         z = inputs_x
#         for i in range(en_ly_size):
#             if i == en_ly_size - 1:
#                 break
#             ns = self.hidden_layers[i]
#             z = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(z)
#             z = BatchNormalization(center=True, scale=False)(z)
#             z = Activation(Relu)(z)
#             z = Dropout(self.dropout_rate_small)(z)
#         latent_z_size = self.hidden_layers[i]
#
#         z_mean = Dense(latent_z_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                         name="z_mean")(z)
#         z_log_var = Dense(latent_z_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#                            name="z_log_var")(z)
#         z = Lambda(sampling, output_shape=(latent_z_size,), name='z')([z_mean, z_log_var])
#         encoder_z = Model(inputs_x, [z_mean, z_log_var, z], name='encoder_z')
#
#         latent_inputs_x = Input(shape=(self.latent_xy_size,), name='x')
#         latent_inputs_xz = Input(shape=(latent_z_size,), name='z_x')
#         latent_x = concatenate([latent_inputs_x, latent_inputs_xz], axis=1)
#         # x = Dense(self.latent_xy_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#         #           activation="linear")(latent_inputs_x)
#         #
#         # lz = Dense(self.latent_z_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
#         #           activation="linear")(latent_inputs_xz)
#         #
#         # x = concatenate([x, lz], axis=1)
#         x = latent_x
#         for i in range(en_ly_size - 1, 0, -1):
#             ns = self.hidden_layers[i - 1]
#             x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
#             x = BatchNormalization(center=True, scale=False)(x)
#             x = Activation(Relu)(x)
#             x = Dropout(self.dropout_rate_big)(x)
#
#         outputs_x = Dense(self.input_size_x, kernel_regularizer=self.kernel_regularizer,
#                         kernel_initializer=self.initializers, activation="softplus")(x)
#         decoder_x = Model([latent_inputs_x, latent_inputs_xz], outputs_x, name='decoder_x')
#
#         latent_inputs_y = Input(shape=(self.latent_xy_size,), name='y')
#         latent_inputs_yz = Input(shape=(latent_z_size,), name='z_y')
#         latent_y = concatenate([latent_inputs_y, latent_inputs_yz], axis=1)
#         y = latent_y
#         for i in range(en_ly_size - 1, 0, -1):
#             ns = self.hidden_layers[i - 1]
#             y = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(y)
#             y = BatchNormalization(center=True, scale=False)(y)
#             y = Activation(Relu)(y)
#             y = Dropout(self.dropout_rate_big)(y)
#
#         outputs_y = Dense(self.input_size_y, kernel_regularizer=self.kernel_regularizer,
#                             kernel_initializer=self.initializers, activation="softplus", name='outputs_y')(y)
#         decoder_y = Model([latent_inputs_y, latent_inputs_yz], outputs_y, name='decoder_y')
#         # decoder_y = Model(latent_inputs, outputs_y, name='decoder_y')
#         # decoder_x = Model(latent_inputs, outputs_x, name='decoder_x')
#
#         # outputs_x = decoder_x(encoder(inputs_x)[0])
#         # outputs_y = decoder_y(encoder(inputs_x)[0])
#         outputs_x = decoder_x([encoder_hx(inputs_x)[2], encoder_z(inputs_x)[2]])
#         outputs_y = decoder_y([encoder_hy(inputs_y)[2], encoder_z(inputs_x)[2]])
#         VIMCCA = Model([inputs_x, inputs_y], [outputs_x, outputs_y], name='vae_mlp')
#         reconstruction_loss_x = K.sum(mse(inputs_x, outputs_x), axis=-1)*10
#         reconstruction_loss_y = K.sum(mse(inputs_y, outputs_y), axis=-1)*2000
#         # reconstruction_loss_x = mse(inputs_x, outputs_x) * 2000
#         # reconstruction_loss_y = mse(inputs_y, outputs_y)*2000
#         kl_loss_z = -0.5 * K.sum(1 + z_log_var - K.square(z_mean) - K.exp(z_log_var), axis=-1)
#         kl_loss_hx = -0.5 * K.sum(1 + hx_log_var - K.square(hx_mean) - K.exp(hx_log_var), axis=-1)
#         kl_loss_hy = -0.5 * K.sum(1 + hy_log_var - K.square(hy_mean) - K.exp(hy_log_var), axis=-1)
#
#         vae_loss = K.mean(reconstruction_loss_x + reconstruction_loss_y + kl_loss_z + kl_loss_hx +
#                           kl_loss_hy)
#
#         VIMCCA.add_loss(vae_loss)
#         self.VIMCCA = VIMCCA
#         self.encoder_z = encoder_z
#         self.encoder_hx = encoder_hx
#         self.encoder_hy = encoder_hy
#         self.decoder_x = decoder_x
#         self.decoder_y = decoder_y
#
#     def compile(self):
#         self.VIMCCA.compile(optimizer=self.optimizer)
#         self.VIMCCA.summary()
#
#     def train(self, x, y, batch_size=100, epochs=300):
#         # if os.path.isfile(self.path + "vae_weights.h5"):
#         #     self.vae.load_weights(self.path + "vae_weights.h5")
#         # else:
#         #
#         history = self.VIMCCA.fit({'inputs_x': x, 'inputs_y': y}, epochs=epochs, batch_size=batch_size,
#                          validation_split=self.validation_split, shuffle=True)
#             # self.vae.model.save(self.path + "/model_best.h5")
#         return history
#
#     def integrate(self, x, save=True, use_mean=True):
#         [z_mean, z_log_var, z] = self.encoder_z.predict(x)
#         return z_mean
#
#     def integrate_compose(self, x, y, save=True, use_mean=True):
#         [z_mean, z_log_var, z] = self.encoder_z.predict(x)
#         [hx_mean, hx_log_var, hx_z] = self.encoder_hx.predict(x)
#         [hy_mean, hy_log_var, hy_z] = self.encoder_hy.predict(y)
#         # res = concatenate([z_mean, hx_mean, hy_mean],axis=1)
#         # res = np.array(res)
#         return z_mean
#
#     def get_output(self, x, y):
#         [z_mean, z_log_var, z] = self.encoder_z.predict(x)
#         [hx_mean, hx_log_var, hx_z] = self.encoder_hx.predict(x)
#         [hy_mean, hy_log_var, hy_z] = self.encoder_hy.predict(y)
#         output_x = self.decoder_x.predict([hx_z, z])
#         output_y = self.decoder_y.predict([hy_z, z])
#         return output_x, output_y
#
#     def plot_loss(self, history):
#         print(len(history.history['loss']))
#         plt.plot(history.history['loss'])
#         plt.show()


def fit_integration(adata_x, adata_y, hidden_layers=[128,64,32,5], epochs=30,weight=5,
                    sparse_x=False, sparse_y=False, batch_size=128):
    """/
    Build VIMCCA model and fit the data to the model for training.

    Parameters
    ----------
    adata_x: AnnData
        AnnData object for one of the modals.

    adata_y: AnnData
        AnnData object for another of the modals.

    epochs: int, optional (default: 200)
        Number of epochs to train the model. An epoch is an iteration over the entire x and y data provided.

    batch_size: int or None, optional (default: 128)
        Number of samples per gradient update. If unspecified, batch_size will default to 128.

    weight: double, optional (default: 5.0)
        The weights of the reconstruction loss for the second modality data.

    sparse_x: bool, optional (default: False)
        If True, Matrix X in the AnnData object is stored as a sparse matrix.

    sparse_y: bool, optional (default: False)
        If True, Matrix Y in the AnnData object is stored as a sparse matrix.

    hidden_layers: list of integers, (default: [128,64,32,5])
        Number of hidden layer neurons in the model.

    Returns
    ----------
        :class:`~Numpy array(s)`
            z
    """

    if sparse_x:
        x = adata_x.X.A
    else:
        x = adata_x.X
    if sparse_y:
        y = adata_y.X.A
    else:
        y = adata_y.X
    net = VIMCCA(input_size_x=x.shape[1], inputs_size_y=y.shape[1], hidden_layers=hidden_layers,weight=weight)
    net.build()
    net.compile()
    his = net.train(x, y, epochs=epochs, batch_size=128)
    z = net.integrate(x)
    return z


import numpy as np
import pandas as pd
from anndata import AnnData
import scanpy as sc
import os
import os.path
import datetime
import tensorflow as tf
from tensorflow.keras.layers import Input, Dense, Activation, BatchNormalization, Dropout, Lambda, Concatenate, concatenate, Add
from tensorflow.keras.models import Model, Sequential, load_model
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint, EarlyStopping, ReduceLROnPlateau
from tensorflow.keras import backend as K
from tensorflow.keras import optimizers, regularizers, initializers
from tensorflow.keras.utils import plot_model
from tensorflow.keras.losses import mse, binary_crossentropy
from scipy.sparse import csr_matrix

def sampling(args):
    z_mean, z_log_var = args
    batch = K.shape(z_mean)[0]
    dim = K.int_shape(z_mean)[1]
    epsilon = K.random_normal(shape=(batch, dim))
    return z_mean + K.exp(0.5 * z_log_var) * epsilon


MeanAct = lambda x: tf.clip_by_value(K.exp(x), 1e-5, 1e5)


class VAE:
    def __init__(self, input_size, path="./",
                 y_size=10,
                 method="lognorm",
                 validation_split=0.0,
                 patience_es=200,
                 patience_lr=100,
                 activation="softplus",
                 lambda_regulizer=2.0,
                 initializers="glorot_uniform",
                 dropout_rate=0.01,
                 hidden_layers=[128, 64, 32],
                 l1_l2=(0.0, 0.0),
                 deterministic=False,
                 save=True):
        self.input_size = input_size
        self.vae = None
        self.inputs = None
        self.outputs = None
        self.path = path
        self.lambda_regulizer = lambda_regulizer
        self.initializers = initializers
        self.method = method
        self.optimizer = optimizers.Adam(lr=0.01)
        self.y_size = y_size
        self.dropout_rate = dropout_rate
        self.hidden_layers = hidden_layers
        self.kernel_regularizer = regularizers.l1_l2(l1=l1_l2[0], l2=l1_l2[1])
        self.validation_split = validation_split
        self.activation = activation
        self.deterministic = deterministic
        self.save = save
        callbacks = []
        checkpointer = ModelCheckpoint(filepath=path + "model_{epoch:04d}.h5", verbose=1, save_best_only=False,
                                       save_weights_only=False, period=100)
        reduce_lr = ReduceLROnPlateau(monitor='loss', factor=0.8, patience=patience_lr, min_lr=0.0001)
        early_stop = EarlyStopping(monitor='loss', patience=patience_es)
        tensor_board = TensorBoard(log_dir=path + "logs" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S/"))
        callbacks.append(checkpointer)
        callbacks.append(reduce_lr)
        callbacks.append(early_stop)
        callbacks.append(tensor_board)
        self.callbacks = callbacks

    def build(self):
        # build encoder
        Relu = "relu"
        inputs = Input(shape=(self.input_size,), name='encoder_input')
        x = Dense(128, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  bias_initializer='zeros', name='en_hidden_layer_x1')(inputs)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)
        x = Dense(64, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  name='en_hidden_layer_x2')(x)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)

        z_mean = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                       name="encoder_mean")(x)
        z_log_var = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                          name="encoder_log_var")(x)
        z = Lambda(sampling, output_shape=(32,), name='hidden_var_z')([z_mean, z_log_var])
        encoder = Model(inputs, [z_mean, z_log_var, z], name='encoder_mlp')

        latent_inputs = Input(shape=(32,), name='z_sampling')
        x = Dense(64, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(latent_inputs)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)
        x = Dense(128, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)
        if self.method == "count":
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="linear")(x)
        elif self.method == "qqnorm":
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="linear")(x)
        else:
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="softplus")(x)
        decoder = Model(latent_inputs, outputs, name='decoder_mlp')
        if self.deterministic:
            outputs = decoder(encoder(inputs)[0])
        else:
            outputs = decoder(encoder(inputs)[2])
        vae = Model(inputs, outputs, name='vae_mlp')
        reconstruction_loss = mse(inputs, outputs)
        reconstruction_loss *= self.input_size
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = -0.5 * K.sum(kl_loss, axis=-1)
        vae_loss = K.mean(reconstruction_loss + kl_loss)
        vae.add_loss(vae_loss)
        self.vae = vae
        self.encoder = encoder
        self.decoder = decoder

    def compile(self):
        self.vae.compile(optimizer=self.optimizer)
        self.vae.summary()

    def train(self, adata, batch_size=256, epochs=300):
        if os.path.isfile(self.path + "model.h5"):
            # self.vae=load_model(self.path+"model.h5")
            self.vae.load_weights(self.path + "model.h5")
        else:
            self.vae.fit(adata.X, epochs=epochs, batch_size=batch_size, callbacks=self.callbacks,
                         validation_split=self.validation_split, shuffle=True)

    def integrate(self, xadata, save=True, use_mean=True):
        [z_mean, z_log_var, z_batch] = self.encoder.predict(xadata.X)
        if use_mean:
            y_mean = self.decoder.predict(z_mean)
        else:
            y_mean = self.decoder.predict(z_batch)
        yadata = AnnData(X=y_mean, obs=xadata.obs, var=xadata.var)
        yadata.raw = AnnData(X=xadata.raw.X, var=xadata.raw.var)
        if save:
            yadata.write(self.path + "output.h5ad")
        yadata.obsm['X_vipcca'] = z_mean
        return yadata


class CVAE(VAE):
    def __init__(self, batches=16, batches2=8, **kwargs):
        super().__init__(**kwargs)
        self.batches = batches
        self.batches2 = batches2

    def build(self):
        Relu = "relu"
        inputs = Input(shape=(self.input_size,), name='encoder_input')
        inputs_batch = Input(shape=(self.batches,), name='batch_input1')
        y = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(
            inputs_batch)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        y = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(y)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        x = concatenate([inputs, y])

        en_ly_size = len(self.hidden_layers)
        for i in range(en_ly_size):
            if i == en_ly_size - 1:
                break
            ns = self.hidden_layers[i]
            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate)(x)
        ns = self.hidden_layers[i]
        z_mean = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                       name="encoder_mean")(x)
        z_log_var = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                          name="encoder_log_var")(x)
        z = Lambda(sampling, output_shape=(ns,), name='hidden_var_z')([z_mean, z_log_var])
        encoder = Model([inputs, inputs_batch], [z_mean, z_log_var, z], name='encoder_mlp')

        latent_inputs = Input(shape=(ns,), name='z_sampling')
        inputs_batch2 = Input(shape=(self.batches2,), name='batch_input2')
        y = Dense(self.batches2, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(
            inputs_batch2)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        y = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(y)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        v = concatenate([latent_inputs, y])
        x = latent_inputs
        for i in range(en_ly_size - 1, 0, -1):
            ns = self.hidden_layers[i - 1]
            v = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(v)
            v = BatchNormalization(center=True, scale=False)(v)
            v = Activation(Relu)(v)
            v = Dropout(self.dropout_rate)(v)

            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate)(x)

        v = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation=self.activation)(v)
        x = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation="hard_sigmoid")(x)
        x = Add()([v, x])
        outputs = Activation(Relu)(x)
        decoder = Model([latent_inputs, inputs_batch2], outputs, name='decoder_mlp')

        if self.deterministic:
            outputs = decoder([encoder([inputs, inputs_batch])[0], inputs_batch2])
        else:
            outputs = decoder([encoder([inputs, inputs_batch])[2], inputs_batch2])
        vae = Model([inputs, inputs_batch, inputs_batch2], outputs, name='vae_mlp')
        #
        noise = tf.math.subtract(inputs, outputs)
        var = tf.math.reduce_variance(noise)
        reconstruction_loss = mse(inputs, outputs)
        reconstruction_loss *= (0.5 * self.input_size) / var
        reconstruction_loss += (0.5 * self.input_size) * tf.math.log(var)
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = -0.5 * K.sum(kl_loss, axis=-1)
        vae_loss = K.mean(2 * reconstruction_loss + self.lambda_regulizer * kl_loss)
        vae.add_loss(vae_loss)
        self.vae = vae
        self.encoder = encoder
        self.decoder = decoder

    def train(self, adata, batch_size=256, epochs=300, model_file=None):
        if model_file is not None:
            # self.vae=load_model(self.path+model_file)
            self.vae.load_weights(self.path + model_file)
        else:
            self.vae.fit([adata.X.A, adata.obsm['X_batch'], adata.obsm['X_batch2']], epochs=epochs, batch_size=batch_size,
                         callbacks=self.callbacks, validation_split=self.validation_split, shuffle=True)
            self.vae.save(self.path + "model.h5")

    def integrate(self, xadata, save=True, use_mean=True):
        [z_mean, z_log_var, z_batch] = self.encoder.predict([xadata.X.A, xadata.obsm['X_batch']])
        if use_mean:
            z_samples = z_mean
        else:
            z_samples = z_batch

        y_mean = self.decoder.predict([z_samples, xadata.obsm['X_batch2']])
        yadata = AnnData(X=y_mean, obs=xadata.obs, var=xadata.var)
        yadata.raw = xadata.copy()
        yadata.obsm['X_vipcca'] = z_mean

        i_mean = self.decoder.predict([z_samples, np.tile(xadata.obsm['X_batch2'][1], (xadata.shape[0], 1))])
        iadata = AnnData(X=i_mean, obs=xadata.obs, var=xadata.var)
        i_mean[i_mean < 0.1] = 0
        iadata.raw = AnnData(X=csr_matrix(i_mean), obs=xadata.obs, var=xadata.var)
        iadata.obsm['X_vipcca'] = z_mean

        if save:
            iadata.write(self.path + "output.h5ad")
        return yadata


class CVAE2(VAE):
    def __init__(self, batches=2, **kwargs):
        super().__init__(**kwargs)
        self.batches = batches

    def build(self):
        Relu = "relu"
        inputs = Input(shape=(self.input_size,), name='encoder_input')
        en_ly_size = len(self.hidden_layers)
        for i in range(en_ly_size):
            ns = self.hidden_layers[i]
            if i == 0:
                x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(inputs)
            elif i == en_ly_size - 1:
                break
            else:
                x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate)(x)
        ns = self.hidden_layers[i]
        z_mean = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                       name="encoder_mean")(x)
        z_log_var = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                          name="encoder_log_var")(x)
        z = Lambda(sampling, output_shape=(ns,), name='hidden_var_z')([z_mean, z_log_var])
        encoder = Model([inputs], [z_mean, z_log_var, z], name='encoder_mlp')

        latent_inputs = Input(shape=(ns,), name='z_sampling')
        inputs_batch = Input(shape=(self.batches,), name='batch_input')
        y = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(
            inputs_batch)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        y = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(y)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        v = concatenate([latent_inputs, y])
        x = latent_inputs
        for i in range(en_ly_size - 1, 0, -1):
            ns = self.hidden_layers[i - 1]
            v = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(v)
            v = BatchNormalization(center=True, scale=False)(v)
            v = Activation(Relu)(v)
            v = Dropout(self.dropout_rate)(v)

            x = Dense(ns, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
            x = BatchNormalization(center=True, scale=False)(x)
            x = Activation(Relu)(x)
            x = Dropout(self.dropout_rate)(x)

        v = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation=self.activation)(v)
        x = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  activation="hard_sigmoid")(x)
        x = Add()([v, x])
        outputs = Activation(Relu)(x)
        decoder = Model([latent_inputs, inputs_batch], outputs, name='decoder_mlp')

        if self.deterministic:
            outputs = decoder([encoder([inputs])[0], inputs_batch])
        else:
            outputs = decoder([encoder([inputs])[2], inputs_batch])
        vae = Model([inputs, inputs_batch], outputs, name='vae_mlp')

        noise = tf.math.subtract(inputs, outputs)
        var = tf.math.reduce_variance(noise)
        reconstruction_loss = mse(inputs, outputs)
        reconstruction_loss *= (0.5 * self.input_size) / var
        reconstruction_loss += (0.5 * self.input_size) * tf.math.log(var)
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = -0.5 * K.sum(kl_loss, axis=-1)
        vae_loss = K.mean(2 * reconstruction_loss + self.lambda_regulizer * kl_loss)
        vae.add_loss(vae_loss)
        self.vae = vae
        self.encoder = encoder
        self.decoder = decoder

    def train(self, adata, batch_size=256, epochs=300, model_file=None):
        if model_file is not None:
            # self.vae=load_model(self.path+model_file)
            self.vae.load_weights(self.path + model_file)
        else:
            self.vae.fit([adata.X, adata.obsm['X_batch']], epochs=epochs, batch_size=batch_size,
                         callbacks=self.callbacks, validation_split=self.validation_split, shuffle=True)
            self.vae.save(self.path + "model.h5")

    def integrate(self, xadata, use_mean=True, return_corrected_y=True):
        [z_mean, z_log_var, z_batch] = self.encoder.predict([xadata.X])
        if use_mean:
            z_samples = z_mean
        else:
            z_samples = z_batch

        if return_corrected_y:
            i_mean = self.decoder.predict([z_samples, np.tile(xadata.obsm['X_batch'][1], (xadata.shape[0], 1))])
            iadata = AnnData(X=i_mean, obs=xadata.obs, var=xadata.var)
            i_mean[i_mean < 0.1] = 0
            iadata.raw = AnnData(X=csr_matrix(i_mean), obs=xadata.obs, var=xadata.var)
            iadata.obsm['X_vipcca'] = z_mean
            iadata.write(self.path + "output.h5ad")
        else:
            i_mean = self.decoder.predict([z_samples, xadata.obsm['X_batch']])
            iadata = AnnData(X=i_mean, obs=xadata.obs, var=xadata.var)
            i_mean[i_mean < 0.1] = 0
            iadata.raw = AnnData(X=csr_matrix(i_mean), obs=xadata.obs, var=xadata.var)
            iadata.obsm['X_vipcca'] = z_mean
            iadata.write(self.path + "output_dsy.h5ad")
        return iadata


class CVAE3(VAE):
    def __init__(self, batches=2, **kwargs):
        super().__init__(**kwargs)
        self.batches = batches

    def build(self):
        Relu = "relu"
        inputs = Input(shape=(self.input_size,), name='encoder_input')
        x = Dense(128, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  bias_initializer='zeros', name='en_hidden_layer_x1')(inputs)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)
        x = Dense(64, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                  name='en_hidden_layer_x2')(x)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)

        z_mean = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                       name="encoder_mean")(x)
        z_log_var = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers,
                          name="encoder_log_var")(x)
        z = Lambda(sampling, output_shape=(32,), name='hidden_var_z')([z_mean, z_log_var])
        encoder = Model([inputs], [z_mean, z_log_var, z], name='encoder_mlp')

        latent_inputs = Input(shape=(32,), name='z_sampling')
        inputs_batch = Input(shape=(self.batches,), name='batch_input')
        x = Dense(self.batches, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(
            inputs_batch)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        y = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(latent_inputs)
        y = BatchNormalization(center=True, scale=False)(y)
        y = Activation(Relu)(y)
        xy = concatenate([y, x])
        z = Dense(32, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(xy)
        z = BatchNormalization(center=True, scale=False)(z)
        z = Activation(Relu)(z)
        z_mean_discriminator = Dense(32, kernel_regularizer=self.kernel_regularizer,
                                     kernel_initializer=self.initializers)(z)

        translator = Model([latent_inputs, inputs_batch], z_mean_discriminator, name="translator_mlp")

        z_prime_input = Input(shape=(32,), name='discriminator')
        x = Dense(64, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(z_prime_input)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)
        x = Dense(128, kernel_regularizer=self.kernel_regularizer, kernel_initializer=self.initializers)(x)
        x = BatchNormalization(center=True, scale=False)(x)
        x = Activation(Relu)(x)
        x = Dropout(self.dropout_rate)(x)

        if self.method == "count":
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="linear")(x)
        elif self.method == "qqnorm":
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation="linear")(x)
        else:
            outputs = Dense(self.input_size, kernel_regularizer=self.kernel_regularizer,
                            kernel_initializer=self.initializers, activation=self.activation)(x)
        decoder = Model(z_prime_input, outputs, name='decoder_mlp')

        if self.deterministic:
            outputs = decoder(translator([encoder([inputs])[0], inputs_batch]))
        else:
            outputs = decoder(translator([encoder([inputs])[2], inputs_batch]))
        vae = Model([inputs, inputs_batch], outputs, name='cvae_mlp')
        reconstruction_loss = mse(inputs, outputs)
        reconstruction_loss *= self.input_size
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = -0.5 * K.sum(kl_loss, axis=-1)
        vae_loss = K.mean(reconstruction_loss + self.lambda_regulizer * kl_loss)
        vae.add_loss(vae_loss)
        self.vae = vae
        self.encoder = encoder
        self.decoder = decoder
        self.translator = translator

    def train(self, adata, batch_size=256, epochs=300, model_file=None):
        if model_file is not None:
            if os.path.exists(self.path + model_file):
                self.vae.load_weights(self.path + model_file)
            else:
                raise ValueError('The specified model file does not exist.')
        else:
            self.vae.fit([adata.X, adata.obsm['X_batch']], epochs=epochs, batch_size=batch_size,
                         callbacks=self.callbacks, validation_split=self.validation_split, shuffle=True)
            self.vae.save(self.path + "model.h5")

    def integrate(self, xadata, save=True, use_mean=True):
        [z_mean, z_log_var, z_batch] = self.encoder.predict([xadata.X])
        if use_mean:
            z_samples = z_mean
        else:
            z_samples = z_batch
        z_mean_prime = self.translator.predict([z_samples, xadata.obsm['X_batch']])

        y_mean = self.decoder.predict([z_mean_prime])
        yadata = AnnData(X=y_mean, obs=xadata.obs, var=xadata.var)
        yadata.raw = xadata.copy()
        yadata.obsm['X_vipcca'] = z_mean
        yadata.obsm['X_vipcca'] = z_mean_prime
        p_mean = self.decoder.predict([z_mean_prime])
        padata = AnnData(X=y_mean, obs=xadata.obs, var=xadata.var)
        padata.raw = xadata.copy()
        padata.obsm['X_vipcca'] = z_mean
        padata.obsm['X_vipcca'] = z_mean_prime
        if save:
            yadata.write(self.path + "output.h5ad")
            padata.write(self.path + "integrated.h5ad")
        return yadata


class Conf(object):
    pass


class VIPCCA(object):
    """
    Initialize VIPCCA object

    Parameters

    ----------

    patience_es: int, optional (default: 50)
        number of epochs with no improvement after which training will be stopped.

    patience_lr: int, optional (default: 25)
        number of epochs with no improvement after which learning rate will be reduced.

    epochs: int, optional (default: 1000)
        Number of epochs to train the model. An epoch is an iteration over the entire x and y data provided.

    res_path: string, (default: None)
        Folder path to save model training results model.h5 and output data adata.h5ad.

    split_by: string, optional (default: '_batch')
        the obsm_name of obsm used to distinguish different batches.

    method: string, optional (default: 'lognorm')
        the normalization method for input data, one of {"qqnorm","count", other}.

    batch_input_size: int, optional (default: 128)
        the length of the batch vector that concatenate with the input layer.

    batch_input_size2: int, optional (default: 16)
        the length of the batch vector that concatenate with the latent layer.

    activation: string, optional (default: "softplus")
        the activation function of hidden layers.

    dropout_rate: double, optional (default: 0.01)
        the dropout rate of hidden layers.

    hidden_layers: list, optional (default: [128,64,32,16])
        Number of hidden layer neurons in the model

    lambda_regulizer: double, optional (default: 5.0)
        The coefficient multiplied by KL_loss

    initializer: string, optional (default: "glorot_uniform")
        Regularizer function applied to the kernel weights matrix.

    l1_l2: tuple, optional (default: (0.0, 0.0))
        [L1 regularization factor, L2 regularization factor].

    mode: string,  optional (default: 'CVAE')
        one of {"CVAE", "CVAE2", "CVAE3"}

    model_file: string, optional (default: None)
        The file name of the trained model, the default is None

    save: bool, optional (default: True)
        If true, save output adata file.

    """

    def __init__(self,
                 adata_all=None,
                 patience_es=50,
                 patience_lr=25,
                 epochs=1000,
                 res_path=None,
                 split_by="_batch",
                 method="lognorm",
                 hvg=True,
                 batch_input_size=128,
                 batch_input_size2=16,
                 activation="softplus",
                 dropout_rate=0.01,
                 hidden_layers=[128, 64, 32, 16],
                 lambda_regulizer=5.0,
                 initializer="glorot_uniform",
                 l1_l2=(0.0, 0.0),
                 mode="CVAE",
                 model_file=None,
                 save=True):
        self.conf = Conf()
        self.conf.adata_all = adata_all
        self.conf.res_path = res_path
        self.conf.patience_es = patience_es
        self.conf.patience_lr = patience_lr
        self.conf.epochs = epochs
        self.conf.split_by = split_by
        self.conf.method = method
        self.conf.hvg = hvg
        self.conf.batch_input_size = batch_input_size
        self.conf.batch_input_size2 = batch_input_size2
        self.conf.dropout_rate = dropout_rate
        self.conf.hidden_layers = hidden_layers
        self.conf.lambda_regulizer = lambda_regulizer
        self.conf.initializer = initializer
        self.conf.activation = activation
        self.conf.model_file = model_file
        self.conf.l1_l2 = l1_l2
        self.conf.mode = mode
        self.conf.save = save
        self.vipcca_preprocessing()

    def vipcca_preprocessing(self):
        """
        Generate the required random batch id for the VIPCCA model

        """
        batch_int = self.conf.adata_all.obs[self.conf.split_by].astype("category").cat.codes.values
        np.random.seed(2019)
        batch_dic = np.random.randint(10, size=(np.max(batch_int) + 1, self.conf.batch_input_size))
        X_batch = np.zeros((len(batch_int), self.conf.batch_input_size))
        batch_dic2 = np.random.randint(10, size=(np.max(batch_int) + 1, self.conf.batch_input_size2))
        X_batch2 = np.zeros((len(batch_int), self.conf.batch_input_size2))
        for i in range(len(batch_int)):
            X_batch[i, :] = batch_dic[batch_int[i], :]
            X_batch2[i, :] = batch_dic2[batch_int[i], :]
        self.conf.adata_all.obsm["X_batch"] = X_batch
        self.conf.adata_all.obsm["X_batch2"] = X_batch2


    def build(self):
        """
        build VIPCCA model
        """
        if self.conf.mode == "CVAE":
            net = CVAE(input_size=self.conf.adata_all.shape[1],
                       path=self.conf.res_path,
                       batches=self.conf.batch_input_size,
                       batches2=self.conf.batch_input_size2,
                       patience_es=self.conf.patience_es,
                       patience_lr=self.conf.patience_lr,
                       activation=self.conf.activation,
                       lambda_regulizer=self.conf.lambda_regulizer,
                       hidden_layers=self.conf.hidden_layers,
                       initializers=self.conf.initializer,
                       dropout_rate=self.conf.dropout_rate,
                       l1_l2=self.conf.l1_l2,
                       method=self.conf.method)
        elif self.conf.mode == "CVAE2":
            net = CVAE2(input_size=self.conf.adata_all.shape[1],
                        path=self.conf.res_path,
                        batches=self.conf.batch_input_size,
                        patience_es=self.conf.patience_es,
                        patience_lr=self.conf.patience_lr,
                        activation=self.conf.activation,
                        lambda_regulizer=self.conf.lambda_regulizer,
                        hidden_layers=self.conf.hidden_layers,
                        initializers=self.conf.initializer,
                        dropout_rate=self.conf.dropout_rate,
                        l1_l2=self.conf.l1_l2,
                        method=self.conf.method,
                        save=self.conf.save)
        else:
            net = CVAE3(input_size=self.conf.adata_all.shape[1],
                        path=self.conf.res_path,
                        batches=self.conf.batch_input_size,
                        patience_es=self.conf.patience_es,
                        patience_lr=self.conf.patience_lr,
                        activation=self.conf.activation,
                        lambda_regulizer=self.conf.lambda_regulizer,
                        hidden_layers=self.conf.hidden_layers,
                        initializers=self.conf.initializer,
                        dropout_rate=self.conf.dropout_rate,
                        l1_l2=self.conf.l1_l2,
                        method=self.conf.method)
        net.build()
        net.compile()
        self.conf.net = net

    def fit_integrate(self):
        """
        Train the constructed VIPCCA model, integrate the data with the trained model,
        and return the integrated anndata object

        Returns
        -------
        :class:`~anndata.AnnData`
		    adata produced by function self.conf.net.integrate(self.conf.adata_all, save=self.conf.save)
        """
        self.build()
        self.conf.net.train(self.conf.adata_all, epochs=self.conf.epochs, model_file=self.conf.model_file)
        return self.conf.net.integrate(self.conf.adata_all, save=self.conf.save)

    # def plotEmbedding(self, eps=[], group_by="Batch", min_dist=0.5):
    #     self.build()
    #     for epoch in eps:
    #         test_result_path = self.conf.res_path + "e%04d_" % epoch
    #         self.conf.net.vae.load_weights(self.conf.res_path + "weights%04d.h5" % epoch)
    #         adata_transform = self.conf.net.integrate(self.conf.adata_all, save=False)
    #         plotCorrelation(adata_transform.raw.X, adata_transform.X, result_path=test_result_path)
    #         run_embedding(adata_transform, path=test_result_path, method="umap", min_dist=min_dist)
    #         plotEmbedding(adata_transform, path=test_result_path, method='umap', group_by=group_by,
    #                       legend_loc="right margin")

#
# if __name__ == "__main__":
#     parser = argparse.ArgumentParser()
#     parser.add_argument('--epochs', type=int, default=50)
#     parser.add_argument('--batch_size', type=int, default=100)
#     conf = parser.parse_args()




import scanpy as sc
import pandas as pd
import matplotlib.pyplot as pl
import seaborn as sns
import numpy as np

fontsize=15
params = {'legend.fontsize': fontsize,
          'figure.figsize': (8, 8),
          'figure.dpi': 150,
         'axes.labelsize': fontsize,
         'axes.titlesize':fontsize,
         'xtick.labelsize':fontsize,
         'ytick.labelsize':fontsize}
pl.rcParams.update(params)
#sns.set(style='white', rc={'figure.figsize':(5,5), 'figure.dpi':150})

def run_embedding(adata,path="./",n_neighbors=15, init=None, method='umap', resolution=1.0, min_dist=0.5):
	sc.pp.neighbors(adata,use_rep='X_vipcca')
	sc.tl.louvain(adata, directed=False,resolution =resolution)
	if method in ("umap","all"):
		sc.tl.umap(adata,min_dist=min_dist)
	if method in ("tsne","all"):
		sc.tl.tsne(adata, use_rep="X_vipcca",n_jobs=20)

def plotEmbedding(adata,path,group_by="batch",ncol=5, method="umap",legend_loc=None,frameon=True,legend_fontsize=6,title=""):
	filename=path+"2dplot_"+group_by+"_"+method+".png"
	if method in ("umap","all"):
		sc.pl.umap(adata,frameon =frameon,color=group_by,show=False,use_raw=False,legend_loc=legend_loc,legend_fontsize=legend_fontsize)
		pl.title(title)
		pl.savefig(filename)
		pl.close()
			
	if method in ("tsne","all"):
		sc.pl.tsne(adata,frameon =frameon,color=group_by,show=False,legend_fontsize=legend_fontsize)
		pl.title("")
		pl.legend(loc=3,fontsize=legend_fontsize,mode="expand",bbox_to_anchor=(0.0, 1.01, 1, 0.2),ncol=ncol)
		pl.savefig(filename)
		pl.close()
	
	if method in ("location","all"):
		sc.pl.scatter(adata, x='xcoord',y='ycoord',frameon =frameon,color=group_by,show=False,use_raw=False,legend_loc="right margin",legend_fontsize=legend_fontsize)
		pl.title("")
		pl.savefig(filename)
		pl.close()
		
def plotDEG(adata,path,group_by="louvain",method="wilcoxon"):
	current_path=path+"deg_"+group_by+"_"+method+"_"
	sc.tl.pca(adata)
	sc.tl.dendrogram(adata,groupby=group_by,use_rep ="X_pca")
	sc.tl.rank_genes_groups(adata,groupby=group_by,use_raw=False,n_genes=100,method=method,rankby_abs=False,corr_method="benjamini-hochberg")
	sc.tl.filter_rank_genes_groups(adata, min_fold_change=3)
	sc.pl.rank_genes_groups(adata,show=False)
	pl.savefig(current_path+"rank_genes_groups.png")
	sc.pl.rank_genes_groups(adata, key='rank_genes_groups_filtered',show=False)
	pl.savefig(current_path+"rank_genes_groups_filtered.png")
	sc.pl.rank_genes_groups_dotplot(adata, key='rank_genes_groups_filtered')
	pl.savefig(current_path+"rank_genes_groups_filtered_dotplot.png")
	top_ranked_genes=pd.DataFrame(adata.uns['rank_genes_groups']['names'][range(1)])
	top_ranked_genes_index = pd.Index(top_ranked_genes.values.flatten()).drop_duplicates(keep='first')
	sc.pl.stacked_violin(adata,top_ranked_genes_index,groupby=group_by,use_raw=False,show=False)
	pl.savefig(current_path+"stacked_violin.png")
	sc.pl.heatmap(adata,top_ranked_genes_index,groupby=group_by,use_raw=False,show=False,swap_axes=False)
	pl.savefig(current_path+"headmap.png")
	pl.close()

def plotDEG2(adata,path,key_batch="batch",key_celltype="celltype",method="wilcoxon", mode=None):
	celltypes=adata.obs[key_celltype].cat.categories.values
	# sc.pp.scale(adata)
	for ct in celltypes:
		if np.sum(adata.obs[key_celltype]==ct)<10:
			continue
		adata_sub = adata[adata.obs[key_celltype]==ct]
		sc.tl.rank_genes_groups(adata_sub, groupby=key_batch, use_raw=False, n_genes=adata.shape[1], rankby_abs=False)
		catname=adata_sub.obs[key_batch].cat.categories.values[0]
		pvals_adj=adata_sub.uns['rank_genes_groups']['pvals_adj'][catname]
		logfc=adata_sub.uns['rank_genes_groups']['logfoldchanges'][catname]
		t=np.isnan(logfc)
		logfc=logfc[~t]
		pvals_adj=pvals_adj[~t]
		# f=np.absolute(logfc)<10
		# logfc=logfc[f]
		# pvals_adj=pvals_adj[f]
		indicator_neg=np.logical_and(logfc<0, pvals_adj<1e-50)
		indicator_pos=np.logical_and(logfc>0, pvals_adj<1e-50)
		ndeg=np.sum(indicator_neg)
		pdeg=np.sum(indicator_pos)
		
		cv=np.repeat('k', len(logfc))
		cv[indicator_pos]='r'
		cv[indicator_neg]='b'
		
		pl.scatter(logfc, -np.log10(pvals_adj+1e-300), c=cv, s=1)
		pl.axvline(x=0, linewidth=0.5, color='c')
		from adjustText import adjust_text
		text1=pl.text(-7.5, 305, '%d significant genes'%np.int(ndeg))
		text2=pl.text(2.5, 305, '%d significant genes'%np.int(pdeg))
		adjust_text([text1,text2])
		if mode is not None:
			tt="{} ({})".format(mode,ct)
		pl.title(tt)
		pl.xlabel(r'$log_2(FC)\ (ctr/stim)$')
		pl.ylabel(r'$-log_{10}(FDR\ adjusted\ p-value)$')
		pl.savefig(path+"dge_%s.png"%ct.replace('+','p'))
		pl.close()
		
def runGeoSketch(adata,N=10000,use_rep="X_pca"):
	from geosketch import gs
	sc.tl.pca(adata)
	sketch_index = gs(adata.obsm[use_rep], N, replace=False)
	adata.uns['geosketch']=adata.obs.index[sketch_index]
	subdata = adata[adata.obs.index[sketch_index]]
	return subdata

def plotQQdeg(adata,path,groupby="batch",method="wilcoxon"):
	sc.tl.rank_genes_groups(adata, groupby=groupby, method=method,n_genes=adata.shape[1],rankby_abs=True,use_raw=False)
	result = adata.uns['rank_genes_groups']
	groups = result['names'].dtype.names
	df = pd.DataFrame({group + '_' + key[:1]: result[key][group] for group in groups for key in ['names', 'pvals_adj']})
	df.to_csv(path+"markers_"+groupby+"_"+method+".csv")
	pvals=adata.uns['rank_genes_groups']['pvals_adj']['0']+1e-260


def plotQQdeg2(adata,path,groupby="batch",method="wilcoxon"):
	for c in adata.obs.cell_type.cat.categories.values:
		adatatemp= adata[adata.obs.cell_type==c,:]
		patht=path+"pvals"+c.replace("/","_")
		plotQQdeg(adatatemp,patht)


def plotPrediction(err,result_path):
	err=err[err<1000]
	x=range(len(err))
	pl.scatter(x,err,c='r',s=1)
	pl.savefig(result_path+"square_error.png")
	pl.close()


def plotCorrelation(y,y_pred, save=True, result_path='./', show=True, rnum=1e4, lim=20):
	"""\
	Plot correlation between original data and corrected data

	Parameters
	----------

	y: matrix or csr_matrix
		The original data matrix.

	y_pred: matrix or csr_matrix
		The data matrix integrated by vipcca.

	save: bool, optional (default: True)
		If True, save the figure into result_path.

	result_path: string, optional (default: './')
		The path for saving the figure.

	show: bool, optional (default: True)
		If True, show the figure.

	rnum: double, optional (default: 1e4)
		The number of points you want to sample randomly in the matrix.

	lim: int, optional (default: 20)
		the right parameter of matplotlib.pyplot.xlim(left, right)

	"""
	from scipy.sparse import csr_matrix
	if (isinstance(y, csr_matrix)):
		y = y.toarray()
	rx = np.random.choice(y.shape[0], np.int(rnum), replace=True)
	ry = np.random.choice(y.shape[1], np.int(rnum), replace=True)
	pl.rcParams['figure.figsize'] = (8, 7)
	pl.scatter(y[rx, ry], y_pred[rx, ry], c='r', s=1)
	pl.xlim(-1, lim)
	pl.ylim(-1, lim)
	pl.xlabel('uncorrected_x')
	pl.ylabel('corrected_x')
	if show:
		pl.show()
	if save:
		pl.savefig(result_path+"correlation.png")
	
	
	



import setuptools
from setuptools import setup
from setuptools import Extension, dist, find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
      'scanpy',
      'tensorflow==2.4.0',
      'anndata',
      'scipy',
      'pandas',
      'seaborn',
      'keras==2.3.1',
      'python-igraph',
      'louvain',
      'h5py'
]
setup(name='scbean',
      version='0.5.0',
      description='integration',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/jhu99/scbean',
      author='Jialu Hu',
      author_email='jhu@nwpu.edu.cn',
      license='MIT',
      packages=find_packages(),
      install_requires=install_requires,
      zip_safe=False,
      python_requires='>=3.6',)



import scipy
from scipy.sparse import csr_matrix, csc_matrix
from sklearn.preprocessing import QuantileTransformer
import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
import os
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
matplotlib.use('Agg')
sns.set(style='white', rc={'figure.figsize': (5, 5), 'figure.dpi': 150})


def read_sc_data(input_file, fmt='h5ad', backed=None, transpose=False, sparse=False, delimiter=" ", unique_name=True,
				batch_name=None, var_names="gene_symbols"):

	"""\
	Read single cell dataset

	Parameters
	----------

	input_file : string
		The path of the file to be read.

	fmt : string, optional (default: 'h5ad')
		The file type of the file to be read.

	backed : Union[Literal[‘r’, ‘r+’], bool, None] (default: None)
		If 'r', load AnnData in backed mode instead of fully loading it into memory (memory mode).
		If you want to modify backed attributes of the AnnData object, you need to choose 'r+'.

	transpose: bool, optional (default: False)
		Whether to transpose the read data.

	sparse: bool, optional (default: False)
		Whether the data in the dataset is stored in sparse matrix format.

	delimiter: str,  optional (default: ' ')
		Delimiter that separates data within text file. If None, will split at arbitrary number of white spaces,
		which is different from enforcing splitting at single white space ' '.

	unique_name: bool, optional (default: False)
		If Ture, AnnData object execute var_names_make_unique() and obs_names_make_unique() functions.

	batch_name: string, optional (default: None)
		Batch name of current batch data

	var_names: Literal[‘gene_symbols’, ‘gene_ids’] (default: 'gene_symbols')
		The variables index when the file type is 'mtx'.

	Returns
	-------
	:class:`~anndata.AnnData`
		adata
	"""
	if fmt == '10x_h5':
		adata = sc.read_10x_h5(input_file)
	elif fmt == '10x_mtx':
		adata = sc.read_10x_mtx(input_file, var_names=var_names)
	elif fmt == "mtx":
		adata = sc.read_mtx(input_file)
	elif fmt == 'h5ad':
		adata = sc.read_h5ad(input_file, backed=backed)
	elif fmt == "csv":
		adata = sc.read_csv(input_file)
	elif fmt == "txt":
		adata = sc.read_text(input_file, delimiter=delimiter)
	elif fmt == "tsv":
		adata = sc.read_text(input_file, delimiter="\t")
	else:
		raise ValueError('`format` needs to be \'10x_h5\' or \'10x_mtx\'')
	if transpose:
		adata = adata.transpose()
	if sparse:
		adata.X = csr_matrix(adata.X, dtype='float32')
	if unique_name:
		adata.var_names_make_unique()
		adata.obs_names_make_unique()
	if batch_name is not None:
		adata.obs["_batch"]=batch_name
	return adata


def concatenate_datasets(datasets):

	"""/
	concatenate single cell datasets

	Parameters
	----------
	datasets: list
		List of AnnData objects to be concatenate.

	Returns
	-------
	:class:`~anndata.AnnData`
		adata
	"""
	adata = datasets[0].concatenate(datasets[1])
	for i in range(2, len(datasets)):
		adata = adata.concatenate(datasets[i])
	return adata


def davae_preprocessing(datasets, min_cells=1, min_genes=1, n_top_genes=2000, mt_ratio=0.8, lognorm=True, hvg=True,
						index_unique=None):
	"""\
	Preprocess and merge data sets from different batches

	Parameters
	----------

	datasets: list, optional (default: None)
		the list of anndata objects from different batches

	min_cells: int, optional (default: 1)
		Minimum number of counts required for a cell to pass filtering.

	min_genes: int, optional (default: 1)
		Minimum number of counts required for a gene to pass filtering.

	n_top_genes: int, optional (default: 2000)
		Number of highly-variable genes to keep.

	mt_ratio: double, optional (default: 0.8)
		Maximum proportion of mito genes for a cell to pass filtering.

	lognorm: bool, optional (default: True)
		If True, execute lognorm() function.

	hvg: bool, optional (default: True)
		If True, choose hypervariable genes for AnnData object.

	index_unique: string, optional (default: None)
		Make the index unique by joining the existing index names with the batch category, using
		index_unique='-', for instance. Provide None to keep existing indices.

	Returns
	-------
	:class:`~anndata.AnnData`
		adata
	"""
	if lognorm:
		for i in range(len(datasets)):
			sc.pp.filter_genes(datasets[i], min_cells=min_cells)
			sc.pp.filter_cells(datasets[i], min_genes=min_genes)
			mito_genes = datasets[i].var_names.str.upper().str.startswith('MT-')
			datasets[i].obs['percent_mito'] = np.sum(datasets[i][:, mito_genes].X, axis=1).A1 / np.sum(datasets[i].X,
																									   axis=1).A1
			datasets[i] = datasets[i][datasets[i].obs['percent_mito'] < mt_ratio, :]
			datasets[i] = logNormalization(datasets[i])
	all_features = pd.Index([])
	for i in range(len(datasets)):
		all_features = all_features.union(datasets[i].var_names)
	common_features = all_features
	for i in range(len(datasets)):
		common_features = common_features.intersection(datasets[i].var.index)
	df_var = pd.DataFrame(index=all_features)
	df_var['selected'] = 0
	if hvg:
		for i in range(len(datasets)):
			sc.pp.highly_variable_genes(datasets[i], flavor='seurat', n_top_genes=n_top_genes, inplace=True)
			features = datasets[i].var_names[datasets[i].var['highly_variable']]
			df_var.loc[features] += 1
		df_var = df_var.loc[common_features]
		df_var.sort_values(by="selected", ascending=False, inplace=True)
		selected_features = df_var.index[range(n_top_genes)]
	else:
		selected_features = common_features
	datasets[0] = datasets[0][:, selected_features]
	datasets[1] = datasets[1][:, selected_features]

	datasets[0].obs['loss_weight'] = 1.0
	for i in range(1, len(datasets)):
		datasets[i].obs['loss_weight'] = 0.0

	for i in range(len(datasets)):
		datasets[i].obs['batch_label'] = i

	adata = datasets[0].concatenate(datasets[1], index_unique=index_unique)
	for i in range(2, len(datasets)):
		adata = adata.concatenate(datasets[i], index_unique=index_unique)
	return adata


def spatial_preprocessing(datasets, min_cells=1, min_genes=1, n_top_genes=2000, lognorm=True, hvg=True):
	"""\
	Preprocess and merge two visium datasets from different batches

	Parameters
	----------

	datasets: list, optional (default: None)
		The list of anndata objects from different batches

	min_cells: int, optional (default: 1)
		Minimum number of counts required for a cell to pass filtering.

	min_genes: int, optional (default: 1)
		Minimum number of counts required for a gene to pass filtering.

	n_top_genes: int, optional (default: 2000)
		Number of highly-variable genes to keep.

	lognorm: bool, optional (default: True)
		If True, execute lognorm() function.

	hvg: bool, optional (default: True)
		If True, choose hypervariable genes for AnnData object.

	Returns
	-------
	:class:`~anndata.AnnData`
		adata
	"""
	for i in range(len(datasets)):
		sc.pp.normalize_total(datasets[i], inplace=True)
		sc.pp.log1p(datasets[i])
		sc.pp.highly_variable_genes(datasets[i], flavor="seurat", n_top_genes=n_top_genes, inplace=True)
		datasets[i].obs['loss_weight'] = i
	adata_spatial = datasets[0].concatenate(
		datasets[1],
		batch_key="library_id",
		uns_merge="unique",
		batch_categories=[
			k
			for d in [
				datasets[0].uns["spatial"],
				datasets[1].uns["spatial"],
			]
			for k, v in d.items()
		],
	)
	return adata_spatial


def spatial_rna_preprocessing(adata_spatial, adata_rna, lognorm=True, hvg=True, n_top_genes=2000):
	'''\
	Preprocess and merge visium dataset with scRNA-seq dataset.

	Parameters
	----------

	adata_spatial: AnnData
		AnnData object of visium dataset.

	adata_rna: AnnData
		AnnData object of scRNA-seq dataset.

	lognorm: bool, optional (default: True)
		If True, execute lognorm() function.

	hvg: bool, optional (default: True)
		If True, choose hypervariable genes for AnnData object.

	n_top_genes: int, optional (default: 2000)
		Number of highly-variable genes to keep.
		
	Returns
	-------
	:class:`~anndata.AnnData`
		adata
	'''
	if lognorm:
		sc.pp.log1p(adata_spatial)
		sc.pp.log1p(adata_rna)
		if hvg:
			sc.pp.highly_variable_genes(adata_spatial, n_top_genes=8000)
			sc.pp.highly_variable_genes(adata_rna, n_top_genes=8000)
		adata_spatial = adata_spatial[:, adata_spatial.var.highly_variable]
		adata_rna = adata_rna[:, adata_rna.var.highly_variable]
	adata_spatial.obs['loss_weight'] = 0.0
	adata_rna.obs['loss_weight'] = 1.0
	adata = adata_spatial.concatenate(adata_rna)
	return adata


def generate_batch_code(batch, batch_num):
	batch_code = []
	if batch_num == 2:
		batch_code = to_categorical(batch)
	if batch_num == 3:
		for i in range(batch.shape[0]):
			if batch[i] == 0:
				batch_code.append([1, 1, 0, 1])
			elif batch[i] == 1:
				batch_code.append([0, 0, 1, 1])
			else:
				batch_code.append([0, 1, 1, 0])
		batch_num = batch_num + 1
	else:
		batch_code = to_categorical(batch)
	batch_code = np.array(batch_code)
	return batch_code, batch_num


def add_location(adata):
	adata_loc = AnnData(adata.obs[['xcoord','ycoord']])
	adata = adata.T.concatenate(adata_loc.T, index_unique=None).T
	adata.obs.rename(columns={'xcoord-0':'xcoord','ycoord-0':'ycoord'},inplace=True)
	return adata


def write2mtx(adata, path):
	if not os.path.exists(path):
		os.makedirs(path)
	gn=adata.var
	bc=adata.obs
	gn.to_csv(path+"genes.tsv",sep="\t",index=True,header=False)
	bc.to_csv(path+"annotation.tsv",sep="\t",index=True,header=False)
	bc.to_csv(path+"barcodes.tsv",index=True,header=False,columns=[])
	scipy.io.mmwrite(path+"matrix.mtx", adata.X.transpose().astype(int))


def recipe_vipcca(adata, n_top_genes=2000, ncounts=1e6, min_cells=10, min_genes=10, max_genes=2500,mt_ratio=0.05,lognorm=True,loc=False,filter=True,hvg=True):
	if filter:
			sc.pp.filter_genes(adata,min_cells=min_cells)
			sc.pp.filter_cells(adata,min_genes=min_genes)
			mito_genes = adata.var_names.str.upper().str.startswith('MT-')
			adata.obs['percent_mito'] = np.sum(adata[:, mito_genes].X, axis=1).A1 / np.sum(adata.X, axis=1).A1
			adata = adata[adata.obs['n_genes'] < max_genes, :]
			adata = adata[adata.obs['percent_mito'] < mt_ratio, :]
	if lognorm:
		adata_norm = logNormalization(adata, loc=loc)
	sc.pp.highly_variable_genes(adata_norm, flavor='seurat', n_top_genes =n_top_genes , inplace=True)
	if loc:
		adata_norm.var.loc[['xcoord','ycoord'],'highly_variable']=[True,True]
	features = adata_norm.var_names[adata_norm.var['highly_variable']]
	adata_norm = adata_norm[:,features]
	sc.pp.filter_cells(adata_norm, min_genes=5)
	adata = adata[adata_norm.obs.index,:]
	adata = adata[:,features]
	adata_norm.raw = adata.copy()
	return adata_norm, adata


def preprocessing(datasets, min_cells=1, min_genes=1, n_top_genes=2000, mt_ratio=0.8, lognorm=True, hvg=True,
				index_unique=None):
	"""\
	Preprocess and merge data sets from different batches

	Parameters
	----------

	datasets: list, optional (default: None)
		the list of anndata objects from different batches

	min_cells: int, optional (default: 1)
		Minimum number of counts required for a cell to pass filtering.

	min_genes: int, optional (default: 1)
		Minimum number of counts required for a gene to pass filtering.

	n_top_genes: int, optional (default: 2000)
		Number of highly-variable genes to keep.

	mt_ratio: double, optional (default: 0.8)
		Maximum proportion of mito genes for a cell to pass filtering.

	lognorm: bool, optional (default: True)
		If True, execute lognorm() function.

	hvg: bool, optional (default: True)
		If True, choose hypervariable genes for AnnData object.

	index_unique: string, optional (default: None)
		Make the index unique by joining the existing index names with the batch category, using
		index_unique='-', for instance. Provide None to keep existing indices.

	Returns
	-------
	:class:`~anndata.AnnData`
		adata_norm
	"""
	if lognorm:
		for i in range(len(datasets)):
			sc.pp.filter_genes(datasets[i],min_cells=min_cells)
			sc.pp.filter_cells(datasets[i],min_genes=min_genes)
			mito_genes = datasets[i].var_names.str.upper().str.startswith('MT-')
			datasets[i].obs['percent_mito'] = np.sum(datasets[i][:, mito_genes].X, axis=1).A1 / np.sum(datasets[i].X, axis=1).A1
			datasets[i] = datasets[i][datasets[i].obs['percent_mito'] < mt_ratio, :]
			datasets[i]=logNormalization(datasets[i])
	all_features=pd.Index([])
	for i in range(len(datasets)):
		all_features=all_features.union(datasets[i].var_names)
	common_features=all_features
	for i in range(len(datasets)):
		common_features=common_features.intersection(datasets[i].var.index)
	df_var = pd.DataFrame(index=all_features)
	df_var['selected']=0
	if hvg:
		for i in range(len(datasets)):
			sc.pp.highly_variable_genes(datasets[i], flavor='seurat', n_top_genes =n_top_genes , inplace=True)
			features = datasets[i].var_names[datasets[i].var['highly_variable']]
			df_var.loc[features]+=1
		df_var=df_var.loc[common_features]
		df_var.sort_values(by="selected", ascending=False, inplace=True)
		selected_features=df_var.index[range(n_top_genes)]
	else:
		selected_features=common_features
	for i in range(len(datasets)):
		datasets[i] = datasets[i][:,selected_features]
		if i==0:
			adata=datasets[i]
		else:
			adata=adata.concatenate(datasets[i],index_unique =index_unique)
	return adata


def split_object(adata,by="batch"):
	adata.obs[by]=adata.obs[by].astype("category")
	index_cat = adata.obs[by].cat.categories
	datasets=[]
	for cat in index_cat:
		data=adata[adata.obs[by]==cat,:]
		datasets.append(data)
	return datasets


def scale(adata,max_value=None,use_rep=None):
	if use_rep is None:
		sc.pp.scale(adata,max_value=max_value)
	else:
		maxv=adata.obsm[use_rep].max(0)
		minv=adata.obsm[use_rep].min(0)
		adata.obsm[use_rep]=(adata.obsm[use_rep]-minv)/(maxv-minv)


def rank_normalization(adata,n_quantiles=2000):
	qt = QuantileTransformer(n_quantiles=n_quantiles,output_distribution='normal',ignore_implicit_zeros=False)
	adata.X = qt.fit_transform(adata.X.toarray())


def logNormalization(adata,loc=False):
	adata.obs['n_counts'] = adata.X.sum(axis=1).A1
	adata.obs['size_factor']=adata.obs['n_counts']/1e6
	adata.raw=adata.copy()
	sc.pp.normalize_total(adata,target_sum=1e6)
	if loc:
		adata_loc = add_location(adata)
		adata_loc.obs.rename(columns={'n_counts-0':'n_counts','size_factor-0':'size_factor'},inplace=True)
		sc.pp.log1p(adata_loc)
		return adata_loc
	else:
		sc.pp.log1p(adata)
		return adata


def createCoordination(adata, result_path=None, plot=False, resolution=1):
	xmax=np.ceil(adata.obs.xcoord.max())+resolution
	ymax=np.ceil(adata.obs.ycoord.max())+resolution
	nn=(xmax/resolution).astype("uint16")
	mm=(ymax/resolution).astype("uint16")
	img_size=np.int(np.ceil(np.max([nn,mm])/216)*216)
	mX=np.zeros(shape=(img_size,img_size),dtype="uint16")
	xcoord_img = (adata.obs.xcoord.values/resolution).astype("uint16")
	ycoord_img = (adata.obs.ycoord.values/resolution).astype("uint16")
	mX[ycoord_img,xcoord_img]+=1
	adata.obs["xcoord_img"]=xcoord_img
	adata.obs["ycoord_img"]=ycoord_img
	count = csr_matrix(([1]*len(xcoord_img),(xcoord_img,ycoord_img)),shape=(img_size,img_size),dtype="int")
	adata.obs["overlapped"]=count.toarray()[xcoord_img,ycoord_img]
	adata.X = csc_matrix((adata.X.T.todense()/adata.obs.overlapped.values).T)
	size_factor = np.zeros(shape=(img_size,img_size,1),dtype="float32")
	size_factor[xcoord_img,ycoord_img,0] = adata.obs["size_factor"]
	if plot:
		plt.scatter(xcoord_img,ycoord_img,s=1,c='r')
		plt.title("All cells in one image")
		plt.savefig(result_path+"image_cells.png")
		plt.close()
	return img_size, img_size, size_factor


def generate_img_from_genes(adata,img_size=648,batch_size=32, cols=[6,7]):
	while True:
		for i in range(adata.shape[1]):
			data=adata.X.getcol(i).data
			ind=adata.X.getcol(i).indices
			xx = adata.obs.iloc[ind,cols[0]].values
			yy = adata.obs.iloc[ind,cols[1]].values
			img = csr_matrix((data,(yy,xx)),shape=(img_size,img_size),dtype="f4").toarray().reshape(1,img_size,img_size,1)
			yield (img,img)


def generate_datasets(adata,img_size=648, cols=[6,7], count_only=False):
	n_genes=adata.shape[1]
	x=np.zeros(shape=(n_genes,img_size,img_size,1),dtype="float32")
	x_label=np.zeros_like(x)
	for i in range(n_genes):
		fi=i
		data=adata.X.getcol(fi).data
		ind=adata.X.getcol(fi).indices
		xx = adata.obs.ix[ind, 'xcoord_img'].values
		yy = adata.obs.ix[ind, 'ycoord_img'].values
		## accumalate cells in the same location
		x[i,]=csr_matrix((data,(xx,yy)),shape=(img_size,img_size),dtype="float32").toarray().reshape(img_size,img_size,1)
		if not count_only:
			rawdata=adata.raw.X.getcol(fi).data
			x_label[i,] = csr_matrix((rawdata,(xx,yy)),shape=(img_size,img_size),dtype="float32").toarray().reshape(img_size,img_size,1)
		else:
			x_label=x
	return x, x_label
	
## generate data used for fit_gen()
# x_train, x_train_label, x_val, x_val_label = pp.generate_datasets(adata,img_size=img_size,validation_split=validation_split)
# datagen = ImageDataGenerator()
# generator_train=datagen.flow(x_train, x_train_label, batch_size=batch_size)
# generator_val=datagen.flow(x_val, x_val_label, batch_size=batch_size)
# unet.fit_gen(generator_train, validation_data=generator_val)


def generate_datasets_ext(adata,img_size=648,validation_split=0.2):
	n_genes=adata.shape[1]
	samples_for_train=n_genes-np.int(n_genes*validation_split)
	id_list=np.random.permutation(n_genes)
	x_train=np.zeros((samples_for_train,img_size,img_size,1),dtype="f4")
	x_train_label=np.zeros_like(x_train)
	x_val=np.zeros((n_genes-samples_for_train,img_size,img_size,1),dtype="f4")
	x_val_label=np.zeros_like(x_val)
	p=0
	for i in id_list[range(samples_for_train)]:
		data=adata.X.getcol(i).data
		rawdata=adata.raw.X.getcol(i).data
		ind=adata.X.getcol(i).indices
		xx = adata.obs.iloc[ind,6].values
		yy = adata.obs.iloc[ind,7].values
		x_train[p,] = csr_matrix((data,(yy,xx)),shape=(img_size,img_size),dtype="f4").toarray().reshape(1,img_size,img_size,1)
		x_train_label[p,] = csr_matrix((rawdata,(yy,xx)),shape=(img_size,img_size),dtype="f4").toarray().reshape(1,img_size,img_size,1)
		p+=1
	p=0	
	for i in id_list[np.arange(samples_for_train,n_genes)]:
		data=adata.X.getcol(i).data
		rawdata=adata.raw.X.getcol(i).data
		ind=adata.X.getcol(i).indices
		xx = adata.obs.iloc[ind,6].values
		yy = adata.obs.iloc[ind,7].values
		x_val[p,] = csr_matrix((data,(yy,xx)),shape=(img_size,img_size),dtype="f4").toarray().reshape(1,img_size,img_size,1)
		x_val_label[p,] = csr_matrix((rawdata,(yy,xx)),shape=(img_size,img_size),dtype="f4").toarray().reshape(1,img_size,img_size,1)
		p+=1
	return x_train, x_train_label, x_val, x_val_label

